#!/usr/bin/python
# Software License Agreement (BSD License)
#
# Copyright (c) 2010, Willow Garage, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#  * Neither the name of the Willow Garage nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

##\author Kevin Watts
##\brief Monitors battery status, and send high/low power status messages

from __future__ import with_statement

PKG = 'continuous_ops_executive'
import roslib; roslib.load_manifest(PKG)
import rospy

import threading

from pr2_msgs.msg import PowerState
from continuous_ops_msgs.msg import ChargeStatus

from collections import deque

CHARGE_LEVEL = 95
DISCHARGE_LEVEL = 50

BUFFER_SIZE = 10

# Return true if we are fully charged
# True if we have 5 consecutive messages above charge level
# And last message shows we're still charging
def _check_stuffed(msgs):
    if not len(msgs) >= 5:
        return False

    for i in range(0, 5):
        m = msgs[-1 * i]
        if not m.relative_capacity > CHARGE_LEVEL:
            return False
        
    if msgs[-1].AC_present == 0:
        return False
    
    return True

class BatteryMonitor(object):
    def __init__(self):
        self._mutex = threading.Lock()
        self._state_pub = rospy.Publisher('charge_status', ChargeStatus, latch=True)
        self._msgs = deque()
        self._power_sub = rospy.Subscriber('power_state', PowerState, self._power_cb)

    def _power_cb(self, msg):
        with self._mutex:
            self._msgs.append(msg)
            
            # Clip deque to max length
            while len(self._msgs) > BUFFER_SIZE:
                self._msgs.popleft()

    def update(self):
        with self._mutex:
            if not self._msgs:
                return

            out = ChargeStatus()
            out.charge_status = ChargeStatus.UNPLUGGED
            out.battery_status = ChargeStatus.HAPPY

            last = self._msgs[-1]

            if last.AC_present > 0:
                out.charge_status = ChargeStatus.CHARGING

            # Hungry if we're not charging and below the minimum level
            if last.AC_present == 0 and last.relative_capacity < DISCHARGE_LEVEL:
                out.battery_status = ChargeStatus.HUNGRY
                
            # Check if we should be stuffed
            if _check_stuffed(self._msgs):
                out.battery_status = ChargeStatus.STUFFED

            self._state_pub.publish(out)


if __name__ == '__main__':
    rospy.init_node('battery_monitor')
    bm = BatteryMonitor()

    try:
        my_rate = rospy.Rate(1.0)
        while not rospy.is_shutdown():
            bm.update()
            my_rate.sleep()
    except KeyboardInterrupt, e:
        pass
    except Exception, e:
        import traceback
        traceback.print_exc()

    
